/*-
 * Copyright 2020 Hochschule für Technik Stuttgart
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package de.hft.stuttgart.quality;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertNull;
import static org.junit.jupiter.api.Assertions.assertTrue;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.InputStream;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.util.List;
import java.util.ServiceLoader;

import org.citygml4j.CityGMLContext;
import org.citygml4j.builder.jaxb.CityGMLBuilder;
import org.citygml4j.builder.jaxb.CityGMLBuilderException;
import org.citygml4j.model.citygml.ade.ADEComponent;
import org.citygml4j.model.citygml.ade.ADEException;
import org.citygml4j.model.citygml.ade.binding.ADEContext;
import org.citygml4j.model.citygml.building.Building;
import org.citygml4j.model.citygml.core.AbstractCityObject;
import org.citygml4j.model.citygml.core.CityModel;
import org.citygml4j.model.citygml.core.CityObjectMember;
import org.citygml4j.model.gml.geometry.primitives.DirectPosition;
import org.citygml4j.model.gml.measures.Angle;
import org.citygml4j.model.gml.measures.Length;
import org.citygml4j.model.module.citygml.CityGMLVersion;
import org.citygml4j.xml.io.CityGMLInputFactory;
import org.citygml4j.xml.io.CityGMLOutputFactory;
import org.citygml4j.xml.io.reader.CityGMLReadException;
import org.citygml4j.xml.io.reader.CityGMLReader;
import org.citygml4j.xml.io.reader.FeatureReadMode;
import org.citygml4j.xml.io.writer.CityGMLWriteException;
import org.citygml4j.xml.io.writer.CityGMLWriter;
import org.junit.jupiter.api.Test;

import de.hft.stuttgart.quality.model.AllPolygonsWrongOrientation;
import de.hft.stuttgart.quality.model.AttributeMissing;
import de.hft.stuttgart.quality.model.AttributeWrongValue;
import de.hft.stuttgart.quality.model.ConsecutivePointsSame;
import de.hft.stuttgart.quality.model.Edge;
import de.hft.stuttgart.quality.model.HoleOutside;
import de.hft.stuttgart.quality.model.InnerRingsNested;
import de.hft.stuttgart.quality.model.InteriorDisconnected;
import de.hft.stuttgart.quality.model.IntersectingRings;
import de.hft.stuttgart.quality.model.MultipleConnectedComponents;
import de.hft.stuttgart.quality.model.NonManifoldEdge;
import de.hft.stuttgart.quality.model.NonManifoldVertex;
import de.hft.stuttgart.quality.model.NonPlanarDistancePlane;
import de.hft.stuttgart.quality.model.NonPlanarNormalsDeviation;
import de.hft.stuttgart.quality.model.OrientationRingsSame;
import de.hft.stuttgart.quality.model.PolygonWrongOrientation;
import de.hft.stuttgart.quality.model.RingNotClosed;
import de.hft.stuttgart.quality.model.RingSelfIntersection;
import de.hft.stuttgart.quality.model.SolidNotClosed;
import de.hft.stuttgart.quality.model.TooFewPoints;
import de.hft.stuttgart.quality.model.TooFewPolygons;
import de.hft.stuttgart.quality.model.Validation;
import de.hft.stuttgart.quality.model.ValidationError;
import de.hft.stuttgart.quality.model.ValidationResult;
import de.hft.stuttgart.quality.model.jaxb.Checking;
import de.hft.stuttgart.quality.model.jaxb.Component;
import de.hft.stuttgart.quality.model.jaxb.ErrorId;
import de.hft.stuttgart.quality.model.jaxb.ErrorStatistics;
import de.hft.stuttgart.quality.model.jaxb.FeatureStatistics;
import de.hft.stuttgart.quality.model.jaxb.Parameter;
import de.hft.stuttgart.quality.model.jaxb.Requirement;
import de.hft.stuttgart.quality.model.jaxb.RequirementId;
import de.hft.stuttgart.quality.model.jaxb.ResultType;
import de.hft.stuttgart.quality.model.jaxb.RingSelfIntType;
import de.hft.stuttgart.quality.model.jaxb.Statistics;
import de.hft.stuttgart.quality.model.jaxb.TopLevelFeatureType;
import de.hft.stuttgart.quality.model.jaxb.ValidationPlan;

class QualityAdeTests {

	@Test
	void testReadingConsecutiveSameError()
			throws ADEException, CityGMLBuilderException, CityGMLReadException, FileNotFoundException {
		CityModel cityModel = readCityGml("src/test/resources/SimpleSolid_Error_QualityADE.gml");
		testValidationPlan(cityModel);

		List<CityObjectMember> cityObjectMember = cityModel.getCityObjectMember();
		assertEquals(1, cityObjectMember.size());
		CityObjectMember com = cityObjectMember.get(0);
		AbstractCityObject aco = com.getCityObject();
		assertNotNull(aco);
		assertTrue(aco instanceof Building);
		Building b = (Building) aco;

		List<ADEComponent> gen = b.getGenericApplicationPropertyOfCityObject();
		ADEComponent resultComp = gen.get(0);
		assertTrue(resultComp instanceof ValidationResult);
		ValidationResult result = (ValidationResult) resultComp;
		assertEquals(ResultType.ERROR, result.getResult());
		assertFalse(result.getErrors().isEmpty());
		List<ValidationError> errors = result.getErrors();
		assertEquals(1, errors.size());
		ValidationError validationError = errors.get(0);
		assertTrue(validationError instanceof ConsecutivePointsSame);
		ConsecutivePointsSame err = (ConsecutivePointsSame) validationError;
		assertEquals("_Simple_BD.1_PG.6_LR.1", err.getLinearRingId());
		Double x1 = err.getVertex1().getValue().get(0);
		assertEquals(11.5, x1, 0.0000001);
		Double x2 = err.getVertex2().getValue().get(0);
		assertEquals(11.5, x2, 0.0000001);
	}

	@Test
	void testReadingTooFewPointsError()
			throws ADEException, CityGMLBuilderException, CityGMLReadException, FileNotFoundException {
		CityModel cityModel = readCityGml("src/test/resources/SimpleSolid_TooFewPoints.gml");

		List<CityObjectMember> cityObjectMember = cityModel.getCityObjectMember();
		CityObjectMember com = cityObjectMember.get(0);
		AbstractCityObject aco = com.getCityObject();
		Building b = (Building) aco;

		List<ADEComponent> gen = b.getGenericApplicationPropertyOfCityObject();
		ADEComponent resultComp = gen.get(0);
		ValidationResult result = (ValidationResult) resultComp;
		assertEquals(ResultType.ERROR, result.getResult());
		assertFalse(result.getErrors().isEmpty());
		List<ValidationError> errors = result.getErrors();
		ValidationError validationError = errors.get(0);
		assertTrue(validationError instanceof TooFewPoints);
		TooFewPoints err = (TooFewPoints) validationError;
		assertEquals("_Simple_BD.1_PG.6_LR.1", err.getLinearRingId());
	}

	@Test
	void testReadingRingSelfIntersectionError()
			throws ADEException, CityGMLBuilderException, CityGMLReadException, FileNotFoundException {
		CityModel cityModel = readCityGml("src/test/resources/SimpleSolid_RingSelfIntersection.gml");

		List<CityObjectMember> cityObjectMember = cityModel.getCityObjectMember();
		CityObjectMember com = cityObjectMember.get(0);
		AbstractCityObject aco = com.getCityObject();
		Building b = (Building) aco;

		List<ADEComponent> gen = b.getGenericApplicationPropertyOfCityObject();
		ADEComponent resultComp = gen.get(0);
		ValidationResult result = (ValidationResult) resultComp;
		assertEquals(ResultType.ERROR, result.getResult());
		assertFalse(result.getErrors().isEmpty());
		List<ValidationError> errors = result.getErrors();
		ValidationError validationError = errors.get(0);
		assertTrue(validationError instanceof RingSelfIntersection);
		RingSelfIntersection err = (RingSelfIntersection) validationError;
		assertEquals("_Simple_BD.1_PG.6_LR.1", err.getLinearRingId());
		assertEquals(RingSelfIntType.POINT_TOUCHES_EDGE, err.getType());
		assertNull(err.getEdge2());
		Edge edge1 = err.getEdge1();
		assertEquals(10.0, edge1.getFrom().getValue().get(0), 0.000001);
		assertEquals(11.0, edge1.getTo().getValue().get(0), 0.000001);
		assertEquals(5.0, err.getVertex1().getValue().get(0));
		assertNull(err.getVertex2());
	}

	@Test
	void testReadingRingNotClosedError()
			throws ADEException, CityGMLBuilderException, CityGMLReadException, FileNotFoundException {
		CityModel cityModel = readCityGml("src/test/resources/SimpleSolid_RingNotClosed.gml");

		List<CityObjectMember> cityObjectMember = cityModel.getCityObjectMember();
		CityObjectMember com = cityObjectMember.get(0);
		AbstractCityObject aco = com.getCityObject();
		Building b = (Building) aco;

		List<ADEComponent> gen = b.getGenericApplicationPropertyOfCityObject();
		ADEComponent resultComp = gen.get(0);
		ValidationResult result = (ValidationResult) resultComp;
		assertEquals(ResultType.ERROR, result.getResult());
		assertFalse(result.getErrors().isEmpty());
		List<ValidationError> errors = result.getErrors();
		ValidationError validationError = errors.get(0);
		assertTrue(validationError instanceof RingNotClosed);
		RingNotClosed err = (RingNotClosed) validationError;
		assertEquals("_Simple_BD.1_PG.6_LR.1", err.getLinearRingId());
	}

	@Test
	void testReadingInteriorDisconnectedError()
			throws ADEException, CityGMLBuilderException, CityGMLReadException, FileNotFoundException {
		CityModel cityModel = readCityGml("src/test/resources/SimpleSolid_InteriorDisconnected.gml");

		List<CityObjectMember> cityObjectMember = cityModel.getCityObjectMember();
		CityObjectMember com = cityObjectMember.get(0);
		AbstractCityObject aco = com.getCityObject();
		Building b = (Building) aco;

		List<ADEComponent> gen = b.getGenericApplicationPropertyOfCityObject();
		ADEComponent resultComp = gen.get(0);
		ValidationResult result = (ValidationResult) resultComp;
		assertEquals(ResultType.ERROR, result.getResult());
		assertFalse(result.getErrors().isEmpty());
		List<ValidationError> errors = result.getErrors();
		ValidationError validationError = errors.get(0);

		assertTrue(validationError instanceof InteriorDisconnected);
		InteriorDisconnected err = (InteriorDisconnected) validationError;
		assertEquals("_Simple_BD.1_PG.1", err.getPolygonId());
	}

	@Test
	void testReadingIntersectingRingsError()
			throws ADEException, CityGMLBuilderException, CityGMLReadException, FileNotFoundException {
		CityModel cityModel = readCityGml("src/test/resources/SimpleSolid_IntersectingRings.gml");

		List<CityObjectMember> cityObjectMember = cityModel.getCityObjectMember();
		CityObjectMember com = cityObjectMember.get(0);
		AbstractCityObject aco = com.getCityObject();
		Building b = (Building) aco;

		List<ADEComponent> gen = b.getGenericApplicationPropertyOfCityObject();
		ADEComponent resultComp = gen.get(0);
		ValidationResult result = (ValidationResult) resultComp;
		assertEquals(ResultType.ERROR, result.getResult());
		assertFalse(result.getErrors().isEmpty());
		List<ValidationError> errors = result.getErrors();
		ValidationError validationError = errors.get(0);

		assertTrue(validationError instanceof IntersectingRings);
		IntersectingRings err = (IntersectingRings) validationError;
		assertEquals("_Simple_BD.1_PG.1", err.getPolygonId());
		assertEquals("_Simple_BD.1_PG.1_Ring1", err.getLinearRingId1());
		assertEquals("_Simple_BD.1_PG.1_Ring2", err.getLinearRingId2());
	}

	@Test
	void testReadingNonPlanarPolygonDistancePlaneError()
			throws ADEException, CityGMLBuilderException, CityGMLReadException, FileNotFoundException {
		CityModel cityModel = readCityGml("src/test/resources/SimpleSolid_NonPlanarPolygonDistancePlane.gml");

		List<CityObjectMember> cityObjectMember = cityModel.getCityObjectMember();
		CityObjectMember com = cityObjectMember.get(0);
		AbstractCityObject aco = com.getCityObject();
		Building b = (Building) aco;

		List<ADEComponent> gen = b.getGenericApplicationPropertyOfCityObject();
		ADEComponent resultComp = gen.get(0);
		ValidationResult result = (ValidationResult) resultComp;
		assertEquals(ResultType.ERROR, result.getResult());
		assertFalse(result.getErrors().isEmpty());
		List<ValidationError> errors = result.getErrors();
		ValidationError validationError = errors.get(0);

		assertTrue(validationError instanceof NonPlanarDistancePlane);
		NonPlanarDistancePlane err = (NonPlanarDistancePlane) validationError;
		assertEquals("_Simple_BD.1_PG.1", err.getPolygonId());
		assertEquals(1.0, err.getVertex().getValue().get(0), 0.000001);
		assertEquals(0.001, err.getDistance().getValue(), 0.000001);
		assertEquals("m", err.getDistance().getUom());
	}

	@Test
	void testReadingInnerRingsNestedError()
			throws ADEException, CityGMLBuilderException, CityGMLReadException, FileNotFoundException {
		CityModel cityModel = readCityGml("src/test/resources/SimpleSolid_InnerRingNested.gml");

		List<CityObjectMember> cityObjectMember = cityModel.getCityObjectMember();
		CityObjectMember com = cityObjectMember.get(0);
		AbstractCityObject aco = com.getCityObject();
		Building b = (Building) aco;

		List<ADEComponent> gen = b.getGenericApplicationPropertyOfCityObject();
		ADEComponent resultComp = gen.get(0);
		ValidationResult result = (ValidationResult) resultComp;
		assertEquals(ResultType.ERROR, result.getResult());
		assertFalse(result.getErrors().isEmpty());
		List<ValidationError> errors = result.getErrors();
		ValidationError validationError = errors.get(0);

		assertTrue(validationError instanceof InnerRingsNested);
		InnerRingsNested err = (InnerRingsNested) validationError;
		assertEquals("_Simple_BD.1_PG.1", err.getPolygonId());
		assertEquals("_Simple_BD.1_PG.1_Ring1", err.getLinearRingId1());
		assertEquals("_Simple_BD.1_PG.1_Ring2", err.getLinearRingId2());
	}

	@Test
	void testReadingHoleOutsideError()
			throws ADEException, CityGMLBuilderException, CityGMLReadException, FileNotFoundException {
		CityModel cityModel = readCityGml("src/test/resources/SimpleSolid_HoleOutside.gml");

		List<CityObjectMember> cityObjectMember = cityModel.getCityObjectMember();
		CityObjectMember com = cityObjectMember.get(0);
		AbstractCityObject aco = com.getCityObject();
		Building b = (Building) aco;

		List<ADEComponent> gen = b.getGenericApplicationPropertyOfCityObject();
		ADEComponent resultComp = gen.get(0);
		ValidationResult result = (ValidationResult) resultComp;
		assertEquals(ResultType.ERROR, result.getResult());
		assertFalse(result.getErrors().isEmpty());
		List<ValidationError> errors = result.getErrors();
		ValidationError validationError = errors.get(0);

		assertTrue(validationError instanceof HoleOutside);
		HoleOutside err = (HoleOutside) validationError;
		assertEquals("_Simple_BD.1_PG.1", err.getPolygonId());
		assertEquals("_Simple_BD.1_PG.1_Ring1", err.getLinearRingId());
	}

	@Test
	void testReadingNonPlanarNormalsDeviationError()
			throws ADEException, CityGMLBuilderException, CityGMLReadException, FileNotFoundException {
		CityModel cityModel = readCityGml("src/test/resources/SimpleSolid_NonPlanarPolygonNormalsDeviation.gml");

		List<CityObjectMember> cityObjectMember = cityModel.getCityObjectMember();
		CityObjectMember com = cityObjectMember.get(0);
		AbstractCityObject aco = com.getCityObject();
		Building b = (Building) aco;

		List<ADEComponent> gen = b.getGenericApplicationPropertyOfCityObject();
		ADEComponent resultComp = gen.get(0);
		ValidationResult result = (ValidationResult) resultComp;
		assertEquals(ResultType.ERROR, result.getResult());
		assertFalse(result.getErrors().isEmpty());
		List<ValidationError> errors = result.getErrors();
		ValidationError validationError = errors.get(0);

		assertTrue(validationError instanceof NonPlanarNormalsDeviation);
		NonPlanarNormalsDeviation err = (NonPlanarNormalsDeviation) validationError;
		assertEquals("_Simple_BD.1_PG.1", err.getPolygonId());
		assertEquals(1.0, err.getDeviation().getValue());
		assertEquals("deg", err.getDeviation().getUom());
	}

	@Test
	void testReadingOrientationRingsSameError()
			throws ADEException, CityGMLBuilderException, CityGMLReadException, FileNotFoundException {
		CityModel cityModel = readCityGml("src/test/resources/SimpleSolid_OrientationRingsSame.gml");

		List<CityObjectMember> cityObjectMember = cityModel.getCityObjectMember();
		CityObjectMember com = cityObjectMember.get(0);
		AbstractCityObject aco = com.getCityObject();
		Building b = (Building) aco;

		List<ADEComponent> gen = b.getGenericApplicationPropertyOfCityObject();
		ADEComponent resultComp = gen.get(0);
		ValidationResult result = (ValidationResult) resultComp;
		assertEquals(ResultType.ERROR, result.getResult());
		assertFalse(result.getErrors().isEmpty());
		List<ValidationError> errors = result.getErrors();
		ValidationError validationError = errors.get(0);

		assertTrue(validationError instanceof OrientationRingsSame);
		OrientationRingsSame err = (OrientationRingsSame) validationError;
		assertEquals("_Simple_BD.1_PG.1", err.getPolygonId());
		assertEquals("_Simple_BD.1_PG.1_Ring1", err.getLinearRingId());
	}

	@Test
	void testReadingAllPolygonsWrongOrientationError()
			throws ADEException, CityGMLBuilderException, CityGMLReadException, FileNotFoundException {
		CityModel cityModel = readCityGml("src/test/resources/SimpleSolid_AllPolygonsWrongOrientation.gml");

		List<CityObjectMember> cityObjectMember = cityModel.getCityObjectMember();
		CityObjectMember com = cityObjectMember.get(0);
		AbstractCityObject aco = com.getCityObject();
		Building b = (Building) aco;

		List<ADEComponent> gen = b.getGenericApplicationPropertyOfCityObject();
		ADEComponent resultComp = gen.get(0);
		ValidationResult result = (ValidationResult) resultComp;
		assertEquals(ResultType.ERROR, result.getResult());
		assertFalse(result.getErrors().isEmpty());
		List<ValidationError> errors = result.getErrors();
		ValidationError validationError = errors.get(0);

		assertTrue(validationError instanceof AllPolygonsWrongOrientation);
		AllPolygonsWrongOrientation err = (AllPolygonsWrongOrientation) validationError;
		assertEquals("_Simple_BD.1_Geom1", err.getGeometryId());
	}

	@Test
	void testReadingPolygonWrongOrientationError()
			throws ADEException, CityGMLBuilderException, CityGMLReadException, FileNotFoundException {
		CityModel cityModel = readCityGml("src/test/resources/SimpleSolid_PolygonWrongOrientation.gml");

		List<CityObjectMember> cityObjectMember = cityModel.getCityObjectMember();
		CityObjectMember com = cityObjectMember.get(0);
		AbstractCityObject aco = com.getCityObject();
		Building b = (Building) aco;

		List<ADEComponent> gen = b.getGenericApplicationPropertyOfCityObject();
		ADEComponent resultComp = gen.get(0);
		ValidationResult result = (ValidationResult) resultComp;
		assertEquals(ResultType.ERROR, result.getResult());
		assertFalse(result.getErrors().isEmpty());
		List<ValidationError> errors = result.getErrors();
		ValidationError validationError = errors.get(0);

		assertTrue(validationError instanceof PolygonWrongOrientation);
		PolygonWrongOrientation err = (PolygonWrongOrientation) validationError;
		assertEquals("_Simple_BD.1_Geom1", err.getGeometryId());
		assertFalse(err.getEdges().isEmpty());
		List<Edge> edges = err.getEdges();
		Edge edge = edges.get(0);
		DirectPosition from = edge.getFrom();
		assertEquals(10.0, from.getValue().get(0), 0.000001);
		assertEquals(11.0, edge.getTo().getValue().get(0), 0.000001);
	}

	@Test
	void testReadingNonManifoldVertexError()
			throws ADEException, CityGMLBuilderException, CityGMLReadException, FileNotFoundException {
		CityModel cityModel = readCityGml("src/test/resources/SimpleSolid_NonManifoldVertex.gml");

		List<CityObjectMember> cityObjectMember = cityModel.getCityObjectMember();
		CityObjectMember com = cityObjectMember.get(0);
		AbstractCityObject aco = com.getCityObject();
		Building b = (Building) aco;

		List<ADEComponent> gen = b.getGenericApplicationPropertyOfCityObject();
		ADEComponent resultComp = gen.get(0);
		ValidationResult result = (ValidationResult) resultComp;
		assertEquals(ResultType.ERROR, result.getResult());
		assertFalse(result.getErrors().isEmpty());
		List<ValidationError> errors = result.getErrors();
		ValidationError validationError = errors.get(0);

		assertTrue(validationError instanceof NonManifoldVertex);
		NonManifoldVertex err = (NonManifoldVertex) validationError;
		assertEquals("_Simple_BD.1_Geom1", err.getGeometryId());
		DirectPosition vertex = err.getVertex();
		assertEquals(1.0, vertex.getValue().get(0), 0.000001);
	}

	@Test
	void testReadingNonManifoldEdgeError()
			throws ADEException, CityGMLBuilderException, CityGMLReadException, FileNotFoundException {
		CityModel cityModel = readCityGml("src/test/resources/SimpleSolid_NonManifoldEdge.gml");

		List<CityObjectMember> cityObjectMember = cityModel.getCityObjectMember();
		CityObjectMember com = cityObjectMember.get(0);
		AbstractCityObject aco = com.getCityObject();
		Building b = (Building) aco;

		List<ADEComponent> gen = b.getGenericApplicationPropertyOfCityObject();
		ADEComponent resultComp = gen.get(0);
		ValidationResult result = (ValidationResult) resultComp;
		assertEquals(ResultType.ERROR, result.getResult());
		assertFalse(result.getErrors().isEmpty());
		List<ValidationError> errors = result.getErrors();
		ValidationError validationError = errors.get(0);

		assertTrue(validationError instanceof NonManifoldEdge);
		NonManifoldEdge err = (NonManifoldEdge) validationError;
		assertEquals("_Simple_BD.1_Geom1", err.getGeometryId());
		assertFalse(err.getEdges().isEmpty());
		List<Edge> edges = err.getEdges();
		Edge edge = edges.get(0);
		DirectPosition from = edge.getFrom();
		assertEquals(10.0, from.getValue().get(0), 0.000001);
		assertEquals(11.0, edge.getTo().getValue().get(0), 0.000001);
	}

	@Test
	void testReadingSolidNotClosedError()
			throws ADEException, CityGMLBuilderException, CityGMLReadException, FileNotFoundException {
		CityModel cityModel = readCityGml("src/test/resources/SimpleSolid_SolidNotClosed.gml");

		List<CityObjectMember> cityObjectMember = cityModel.getCityObjectMember();
		CityObjectMember com = cityObjectMember.get(0);
		AbstractCityObject aco = com.getCityObject();
		Building b = (Building) aco;

		List<ADEComponent> gen = b.getGenericApplicationPropertyOfCityObject();
		ADEComponent resultComp = gen.get(0);
		ValidationResult result = (ValidationResult) resultComp;
		assertEquals(ResultType.ERROR, result.getResult());
		assertFalse(result.getErrors().isEmpty());
		List<ValidationError> errors = result.getErrors();
		ValidationError validationError = errors.get(0);

		assertTrue(validationError instanceof SolidNotClosed);
		SolidNotClosed err = (SolidNotClosed) validationError;
		assertEquals("_Simple_BD.1_Geom1", err.getGeometryId());
		assertFalse(err.getEdges().isEmpty());
		List<Edge> edges = err.getEdges();
		Edge edge = edges.get(0);
		DirectPosition from = edge.getFrom();
		assertEquals(10.0, from.getValue().get(0), 0.000001);
		assertEquals(11.0, edge.getTo().getValue().get(0), 0.000001);
	}

	@Test
	void testReadingTooFewPolygonsError()
			throws ADEException, CityGMLBuilderException, CityGMLReadException, FileNotFoundException {
		CityModel cityModel = readCityGml("src/test/resources/SimpleSolid_TooFewPolygons.gml");

		List<CityObjectMember> cityObjectMember = cityModel.getCityObjectMember();
		CityObjectMember com = cityObjectMember.get(0);
		AbstractCityObject aco = com.getCityObject();
		Building b = (Building) aco;

		List<ADEComponent> gen = b.getGenericApplicationPropertyOfCityObject();
		ADEComponent resultComp = gen.get(0);
		ValidationResult result = (ValidationResult) resultComp;
		assertEquals(ResultType.ERROR, result.getResult());
		assertFalse(result.getErrors().isEmpty());
		List<ValidationError> errors = result.getErrors();
		ValidationError validationError = errors.get(0);

		assertTrue(validationError instanceof TooFewPolygons);
		TooFewPolygons err = (TooFewPolygons) validationError;
		assertEquals("_Simple_BD.1_Geom1", err.getGeometryId());
	}

	@Test
	void testReadingMultipleConnectedComponentsError()
			throws ADEException, CityGMLBuilderException, CityGMLReadException, FileNotFoundException {
		CityModel cityModel = readCityGml("src/test/resources/SimpleSolid_MultipleConnectedComponents.gml");

		List<CityObjectMember> cityObjectMember = cityModel.getCityObjectMember();
		CityObjectMember com = cityObjectMember.get(0);
		AbstractCityObject aco = com.getCityObject();
		Building b = (Building) aco;

		List<ADEComponent> gen = b.getGenericApplicationPropertyOfCityObject();
		ADEComponent resultComp = gen.get(0);
		ValidationResult result = (ValidationResult) resultComp;
		assertEquals(ResultType.ERROR, result.getResult());
		assertFalse(result.getErrors().isEmpty());
		List<ValidationError> errors = result.getErrors();
		ValidationError validationError = errors.get(0);

		assertTrue(validationError instanceof MultipleConnectedComponents);
		MultipleConnectedComponents err = (MultipleConnectedComponents) validationError;
		assertEquals("_Simple_BD.1_Geom1", err.getGeometryId());
		assertEquals("_Simple_BD.1_PG.1", err.getComponents().get(0).getPolygonIds().get(0));
		assertEquals("_Simple_BD.1_PG.2", err.getComponents().get(0).getPolygonIds().get(1));
		assertEquals("_Simple_BD.1_PG.3", err.getComponents().get(1).getPolygonIds().get(0));
		assertEquals("_Simple_BD.1_PG.4", err.getComponents().get(1).getPolygonIds().get(1));
	}
	
	@Test
	void testWritingAttributeMissingError()
			throws ADEException, CityGMLBuilderException, CityGMLWriteException, CityGMLReadException {
		CityModel model = new CityModel();
		Building b = new Building();
		model.getCityObjectMember().add(new CityObjectMember(b));

		ValidationResult result = new ValidationResult();
		result.setResult(ResultType.ERROR);
		b.addGenericApplicationPropertyOfCityObject(result);
		AttributeMissing error = new AttributeMissing();
		error.setAttributeName("TestName");
		error.setChildId("TestId");
		error.setGeneric(true);
		result.getErrors().add(error);
		byte[] gml = writeCityGml(model);
		model = readCityGml(gml);

		AbstractCityObject co = model.getCityObjectMember().get(0).getCityObject();
		List<ADEComponent> ades = co.getGenericApplicationPropertyOfCityObject();
		ADEComponent adeComponent = ades.get(0);
		ValidationResult valRes = (ValidationResult) adeComponent;
		assertEquals(ResultType.ERROR, valRes.getResult());

		ValidationError valError = valRes.getErrors().get(0);
		AttributeMissing error2 = (AttributeMissing) valError;
		assertEquals(error.getAttributeName(), error2.getAttributeName());
		assertEquals(error.getChildId(), error2.getChildId());
		assertEquals(error.isGeneric(), error2.isGeneric());
	}
	
	@Test
	void testWritingAttributeWrongValueError()
			throws ADEException, CityGMLBuilderException, CityGMLWriteException, CityGMLReadException {
		CityModel model = new CityModel();
		Building b = new Building();
		model.getCityObjectMember().add(new CityObjectMember(b));

		ValidationResult result = new ValidationResult();
		result.setResult(ResultType.ERROR);
		b.addGenericApplicationPropertyOfCityObject(result);
		AttributeWrongValue error = new AttributeWrongValue();
		error.setAttributeName("TestName");
		error.setChildId("TestId");
		error.setGeneric(true);
		result.getErrors().add(error);
		byte[] gml = writeCityGml(model);
		model = readCityGml(gml);

		AbstractCityObject co = model.getCityObjectMember().get(0).getCityObject();
		List<ADEComponent> ades = co.getGenericApplicationPropertyOfCityObject();
		ADEComponent adeComponent = ades.get(0);
		ValidationResult valRes = (ValidationResult) adeComponent;
		assertEquals(ResultType.ERROR, valRes.getResult());

		ValidationError valError = valRes.getErrors().get(0);
		AttributeWrongValue error2 = (AttributeWrongValue) valError;
		assertEquals(error.getAttributeName(), error2.getAttributeName());
		assertEquals(error.getChildId(), error2.getChildId());
		assertEquals(error.isGeneric(), error2.isGeneric());
	}

	@Test
	void testWritingMultipleConnectedComponentsError()
			throws ADEException, CityGMLBuilderException, CityGMLWriteException, CityGMLReadException {
		CityModel model = new CityModel();
		Building b = new Building();
		model.getCityObjectMember().add(new CityObjectMember(b));

		ValidationResult result = new ValidationResult();
		result.setResult(ResultType.ERROR);
		b.addGenericApplicationPropertyOfCityObject(result);
		MultipleConnectedComponents error = new MultipleConnectedComponents();
		error.setGeometryId("TestGeometry");
		Component c1 = new Component();
		c1.getPolygonIds().add("TestPoly");
		error.getComponents().add(c1);
		result.getErrors().add(error);

		byte[] gml = writeCityGml(model);
		model = readCityGml(gml);

		AbstractCityObject co = model.getCityObjectMember().get(0).getCityObject();
		List<ADEComponent> ades = co.getGenericApplicationPropertyOfCityObject();
		ADEComponent adeComponent = ades.get(0);
		ValidationResult valRes = (ValidationResult) adeComponent;
		assertEquals(ResultType.ERROR, valRes.getResult());

		ValidationError valError = valRes.getErrors().get(0);
		MultipleConnectedComponents error2 = (MultipleConnectedComponents) valError;
		assertEquals(error.getGeometryId(), error2.getGeometryId());
		assertEquals(error.getComponents().get(0).getPolygonIds().get(0),
				error2.getComponents().get(0).getPolygonIds().get(0));
	}

	@Test
	void testWritingTooFewPolygonsError()
			throws ADEException, CityGMLBuilderException, CityGMLWriteException, CityGMLReadException {
		CityModel model = new CityModel();
		Building b = new Building();
		model.getCityObjectMember().add(new CityObjectMember(b));

		ValidationResult result = new ValidationResult();
		result.setResult(ResultType.ERROR);
		b.addGenericApplicationPropertyOfCityObject(result);
		TooFewPolygons error = new TooFewPolygons();
		error.setGeometryId("TestGeometry");
		result.getErrors().add(error);

		byte[] gml = writeCityGml(model);
		model = readCityGml(gml);

		AbstractCityObject co = model.getCityObjectMember().get(0).getCityObject();
		List<ADEComponent> ades = co.getGenericApplicationPropertyOfCityObject();
		ADEComponent adeComponent = ades.get(0);
		ValidationResult valRes = (ValidationResult) adeComponent;
		assertEquals(ResultType.ERROR, valRes.getResult());

		ValidationError valError = valRes.getErrors().get(0);
		TooFewPolygons error2 = (TooFewPolygons) valError;
		assertEquals(error.getGeometryId(), error2.getGeometryId());
	}

	@Test
	void testWritingSolidNotClosedError()
			throws ADEException, CityGMLBuilderException, CityGMLWriteException, CityGMLReadException {
		CityModel model = new CityModel();
		Building b = new Building();
		model.getCityObjectMember().add(new CityObjectMember(b));

		ValidationResult result = new ValidationResult();
		result.setResult(ResultType.ERROR);
		b.addGenericApplicationPropertyOfCityObject(result);
		SolidNotClosed error = new SolidNotClosed();
		error.setGeometryId("TestGeometry");
		DirectPosition from = new DirectPosition();
		from.getValue().add(10.0);
		DirectPosition to = new DirectPosition();
		to.getValue().add(11.0);
		Edge e = new Edge();
		e.setFrom(from);
		e.setTo(to);
		error.getEdges().add(e);
		result.getErrors().add(error);

		byte[] gml = writeCityGml(model);
		model = readCityGml(gml);

		AbstractCityObject co = model.getCityObjectMember().get(0).getCityObject();
		List<ADEComponent> ades = co.getGenericApplicationPropertyOfCityObject();
		ADEComponent adeComponent = ades.get(0);
		ValidationResult valRes = (ValidationResult) adeComponent;
		assertEquals(ResultType.ERROR, valRes.getResult());

		ValidationError valError = valRes.getErrors().get(0);
		SolidNotClosed error2 = (SolidNotClosed) valError;
		assertEquals(error.getGeometryId(), error2.getGeometryId());
		Edge edge = error.getEdges().get(0);
		assertEquals(edge.getFrom().getValue().get(0), e.getFrom().getValue().get(0), 0.00001);
		assertEquals(edge.getTo().getValue().get(0), e.getTo().getValue().get(0), 0.00001);
	}

	@Test
	void testWritingNonManifoldEdgeError()
			throws ADEException, CityGMLBuilderException, CityGMLWriteException, CityGMLReadException {
		CityModel model = new CityModel();
		Building b = new Building();
		model.getCityObjectMember().add(new CityObjectMember(b));

		ValidationResult result = new ValidationResult();
		result.setResult(ResultType.ERROR);
		b.addGenericApplicationPropertyOfCityObject(result);
		NonManifoldEdge error = new NonManifoldEdge();
		error.setGeometryId("TestGeometry");
		DirectPosition from = new DirectPosition();
		from.getValue().add(10.0);
		DirectPosition to = new DirectPosition();
		to.getValue().add(11.0);
		Edge e = new Edge();
		e.setFrom(from);
		e.setTo(to);
		error.getEdges().add(e);
		result.getErrors().add(error);

		byte[] gml = writeCityGml(model);
		model = readCityGml(gml);

		AbstractCityObject co = model.getCityObjectMember().get(0).getCityObject();
		List<ADEComponent> ades = co.getGenericApplicationPropertyOfCityObject();
		ADEComponent adeComponent = ades.get(0);
		ValidationResult valRes = (ValidationResult) adeComponent;
		assertEquals(ResultType.ERROR, valRes.getResult());

		ValidationError valError = valRes.getErrors().get(0);
		NonManifoldEdge error2 = (NonManifoldEdge) valError;
		assertEquals(error.getGeometryId(), error2.getGeometryId());
		Edge edge = error.getEdges().get(0);
		assertEquals(edge.getFrom().getValue().get(0), e.getFrom().getValue().get(0), 0.00001);
		assertEquals(edge.getTo().getValue().get(0), e.getTo().getValue().get(0), 0.00001);
	}

	@Test
	void testWritingNonManifoldVertexError()
			throws ADEException, CityGMLBuilderException, CityGMLWriteException, CityGMLReadException {
		CityModel model = new CityModel();
		Building b = new Building();
		model.getCityObjectMember().add(new CityObjectMember(b));

		ValidationResult result = new ValidationResult();
		result.setResult(ResultType.ERROR);
		b.addGenericApplicationPropertyOfCityObject(result);
		NonManifoldVertex error = new NonManifoldVertex();
		error.setGeometryId("TestGeometry");
		DirectPosition vertex = new DirectPosition();
		vertex.getValue().add(10.0);
		error.setVertex(vertex);
		result.getErrors().add(error);

		byte[] gml = writeCityGml(model);
		model = readCityGml(gml);

		AbstractCityObject co = model.getCityObjectMember().get(0).getCityObject();
		List<ADEComponent> ades = co.getGenericApplicationPropertyOfCityObject();
		ADEComponent adeComponent = ades.get(0);
		ValidationResult valRes = (ValidationResult) adeComponent;
		assertEquals(ResultType.ERROR, valRes.getResult());

		ValidationError valError = valRes.getErrors().get(0);
		NonManifoldVertex error2 = (NonManifoldVertex) valError;
		assertEquals(error.getGeometryId(), error2.getGeometryId());
		assertEquals(error.getVertex().getValue().get(0), error2.getVertex().getValue().get(0));
	}

	@Test
	void testWritingPolygonWrongOrientationError()
			throws ADEException, CityGMLBuilderException, CityGMLWriteException, CityGMLReadException {
		CityModel model = new CityModel();
		Building b = new Building();
		model.getCityObjectMember().add(new CityObjectMember(b));

		ValidationResult result = new ValidationResult();
		result.setResult(ResultType.ERROR);
		b.addGenericApplicationPropertyOfCityObject(result);
		PolygonWrongOrientation error = new PolygonWrongOrientation();
		error.setGeometryId("TestGeometry");
		DirectPosition from = new DirectPosition();
		from.getValue().add(10.0);
		DirectPosition to = new DirectPosition();
		to.getValue().add(11.0);
		Edge e = new Edge();
		e.setFrom(from);
		e.setTo(to);
		error.getEdges().add(e);
		result.getErrors().add(error);

		byte[] gml = writeCityGml(model);
		model = readCityGml(gml);

		AbstractCityObject co = model.getCityObjectMember().get(0).getCityObject();
		List<ADEComponent> ades = co.getGenericApplicationPropertyOfCityObject();
		ADEComponent adeComponent = ades.get(0);
		ValidationResult valRes = (ValidationResult) adeComponent;
		assertEquals(ResultType.ERROR, valRes.getResult());

		ValidationError valError = valRes.getErrors().get(0);
		PolygonWrongOrientation error2 = (PolygonWrongOrientation) valError;
		assertEquals(error.getGeometryId(), error2.getGeometryId());
		Edge edge = error.getEdges().get(0);
		assertEquals(edge.getFrom().getValue().get(0), e.getFrom().getValue().get(0), 0.00001);
		assertEquals(edge.getTo().getValue().get(0), e.getTo().getValue().get(0), 0.00001);
	}

	@Test
	void testWritingAllPolygonsWrongOrientationError()
			throws ADEException, CityGMLBuilderException, CityGMLWriteException, CityGMLReadException {
		CityModel model = new CityModel();
		Building b = new Building();
		model.getCityObjectMember().add(new CityObjectMember(b));

		ValidationResult result = new ValidationResult();
		result.setResult(ResultType.ERROR);
		b.addGenericApplicationPropertyOfCityObject(result);
		AllPolygonsWrongOrientation error = new AllPolygonsWrongOrientation();
		error.setGeometryId("TestGeometry");
		result.getErrors().add(error);

		byte[] gml = writeCityGml(model);
		model = readCityGml(gml);

		AbstractCityObject co = model.getCityObjectMember().get(0).getCityObject();
		List<ADEComponent> ades = co.getGenericApplicationPropertyOfCityObject();
		ADEComponent adeComponent = ades.get(0);
		ValidationResult valRes = (ValidationResult) adeComponent;
		assertEquals(ResultType.ERROR, valRes.getResult());

		ValidationError valError = valRes.getErrors().get(0);
		AllPolygonsWrongOrientation error2 = (AllPolygonsWrongOrientation) valError;
		assertEquals(error.getGeometryId(), error2.getGeometryId());
	}

	@Test
	void testWritingOrientationRingsSameError()
			throws ADEException, CityGMLBuilderException, CityGMLWriteException, CityGMLReadException {
		CityModel model = new CityModel();
		Building b = new Building();
		model.getCityObjectMember().add(new CityObjectMember(b));

		ValidationResult result = new ValidationResult();
		result.setResult(ResultType.ERROR);
		b.addGenericApplicationPropertyOfCityObject(result);
		OrientationRingsSame error = new OrientationRingsSame();
		error.setPolygonId("TestPolygon");
		error.setLinearRingId("TestRing");
		result.getErrors().add(error);

		byte[] gml = writeCityGml(model);
		model = readCityGml(gml);

		AbstractCityObject co = model.getCityObjectMember().get(0).getCityObject();
		List<ADEComponent> ades = co.getGenericApplicationPropertyOfCityObject();
		ADEComponent adeComponent = ades.get(0);
		ValidationResult valRes = (ValidationResult) adeComponent;
		assertEquals(ResultType.ERROR, valRes.getResult());

		ValidationError valError = valRes.getErrors().get(0);
		OrientationRingsSame error2 = (OrientationRingsSame) valError;
		assertEquals(error.getPolygonId(), error2.getPolygonId());
		assertEquals(error.getLinearRingId(), error2.getLinearRingId());
	}

	@Test
	void testWritingNonPlanarNormalsDeviationError()
			throws ADEException, CityGMLBuilderException, CityGMLWriteException, CityGMLReadException {
		CityModel model = new CityModel();
		Building b = new Building();
		model.getCityObjectMember().add(new CityObjectMember(b));

		ValidationResult result = new ValidationResult();
		result.setResult(ResultType.ERROR);
		b.addGenericApplicationPropertyOfCityObject(result);
		NonPlanarNormalsDeviation error = new NonPlanarNormalsDeviation();
		error.setPolygonId("TestPolygon");
		Angle deviation = new Angle();
		deviation.setValue(1.0);
		deviation.setUom("deg");
		error.setDeviation(deviation);
		result.getErrors().add(error);

		byte[] gml = writeCityGml(model);
		model = readCityGml(gml);

		AbstractCityObject co = model.getCityObjectMember().get(0).getCityObject();
		List<ADEComponent> ades = co.getGenericApplicationPropertyOfCityObject();
		ADEComponent adeComponent = ades.get(0);
		ValidationResult valRes = (ValidationResult) adeComponent;
		assertEquals(ResultType.ERROR, valRes.getResult());

		ValidationError valError = valRes.getErrors().get(0);
		NonPlanarNormalsDeviation error2 = (NonPlanarNormalsDeviation) valError;
		assertEquals(error.getPolygonId(), error2.getPolygonId());
		assertEquals(error.getDeviation().getValue(), error2.getDeviation().getValue());
		assertEquals(error.getDeviation().getUom(), error2.getDeviation().getUom());
	}

	@Test
	void testWritingHoleOutsideError()
			throws ADEException, CityGMLBuilderException, CityGMLWriteException, CityGMLReadException {
		CityModel model = new CityModel();
		Building b = new Building();
		model.getCityObjectMember().add(new CityObjectMember(b));

		ValidationResult result = new ValidationResult();
		result.setResult(ResultType.ERROR);
		b.addGenericApplicationPropertyOfCityObject(result);
		HoleOutside error = new HoleOutside();
		error.setPolygonId("TestPolygon");
		error.setLinearRingId("TestRing1");
		result.getErrors().add(error);

		byte[] gml = writeCityGml(model);
		model = readCityGml(gml);

		AbstractCityObject co = model.getCityObjectMember().get(0).getCityObject();
		List<ADEComponent> ades = co.getGenericApplicationPropertyOfCityObject();
		ADEComponent adeComponent = ades.get(0);
		ValidationResult valRes = (ValidationResult) adeComponent;
		assertEquals(ResultType.ERROR, valRes.getResult());

		ValidationError valError = valRes.getErrors().get(0);
		HoleOutside error2 = (HoleOutside) valError;
		assertEquals(error.getPolygonId(), error2.getPolygonId());
		assertEquals(error.getLinearRingId(), error2.getLinearRingId());
	}

	@Test
	void testWritingInnerRingsNestedError()
			throws ADEException, CityGMLBuilderException, CityGMLWriteException, CityGMLReadException {
		CityModel model = new CityModel();
		Building b = new Building();
		model.getCityObjectMember().add(new CityObjectMember(b));

		ValidationResult result = new ValidationResult();
		result.setResult(ResultType.ERROR);
		b.addGenericApplicationPropertyOfCityObject(result);
		InnerRingsNested error = new InnerRingsNested();
		error.setPolygonId("TestPolygon");
		error.setLinearRingId1("TestRing1");
		error.setLinearRingId2("TestRing2");
		result.getErrors().add(error);

		byte[] gml = writeCityGml(model);
		model = readCityGml(gml);

		AbstractCityObject co = model.getCityObjectMember().get(0).getCityObject();
		List<ADEComponent> ades = co.getGenericApplicationPropertyOfCityObject();
		ADEComponent adeComponent = ades.get(0);
		ValidationResult valRes = (ValidationResult) adeComponent;
		assertEquals(ResultType.ERROR, valRes.getResult());

		ValidationError valError = valRes.getErrors().get(0);
		InnerRingsNested error2 = (InnerRingsNested) valError;
		assertEquals(error.getPolygonId(), error2.getPolygonId());
		assertEquals(error.getLinearRingId1(), error2.getLinearRingId1());
		assertEquals(error.getLinearRingId2(), error2.getLinearRingId2());
	}

	@Test
	void testWritingNonPlanarDistancePlaneError()
			throws ADEException, CityGMLBuilderException, CityGMLWriteException, CityGMLReadException {
		CityModel model = new CityModel();
		Building b = new Building();
		model.getCityObjectMember().add(new CityObjectMember(b));

		ValidationResult result = new ValidationResult();
		result.setResult(ResultType.ERROR);
		b.addGenericApplicationPropertyOfCityObject(result);
		NonPlanarDistancePlane error = new NonPlanarDistancePlane();
		error.setPolygonId("TestPolygon");
		DirectPosition vertex = new DirectPosition();
		vertex.getValue().add(1.0);
		error.setVertex(vertex);
		Length distance = new Length(0.1);
		distance.setUom("m");
		error.setDistance(distance);
		result.getErrors().add(error);

		byte[] gml = writeCityGml(model);
		model = readCityGml(gml);

		AbstractCityObject co = model.getCityObjectMember().get(0).getCityObject();
		List<ADEComponent> ades = co.getGenericApplicationPropertyOfCityObject();
		ADEComponent adeComponent = ades.get(0);
		ValidationResult valRes = (ValidationResult) adeComponent;
		assertEquals(ResultType.ERROR, valRes.getResult());

		ValidationError valError = valRes.getErrors().get(0);
		NonPlanarDistancePlane error2 = (NonPlanarDistancePlane) valError;
		assertEquals(error.getPolygonId(), error2.getPolygonId());
		assertEquals(error.getVertex().getValue().get(0), error2.getVertex().getValue().get(0), 0.00001);
		assertEquals(error.getDistance().getValue(), error2.getDistance().getValue(), 0.00001);
	}

	@Test
	void testWritingIntersectingRingsError()
			throws ADEException, CityGMLBuilderException, CityGMLWriteException, CityGMLReadException {
		CityModel model = new CityModel();
		Building b = new Building();
		model.getCityObjectMember().add(new CityObjectMember(b));

		ValidationResult result = new ValidationResult();
		result.setResult(ResultType.ERROR);
		b.addGenericApplicationPropertyOfCityObject(result);
		IntersectingRings error = new IntersectingRings();
		error.setPolygonId("TestPolygon");
		error.setLinearRingId1("TestRing1");
		error.setLinearRingId2("TestRing2");
		result.getErrors().add(error);

		byte[] gml = writeCityGml(model);
		model = readCityGml(gml);

		AbstractCityObject co = model.getCityObjectMember().get(0).getCityObject();
		List<ADEComponent> ades = co.getGenericApplicationPropertyOfCityObject();
		ADEComponent adeComponent = ades.get(0);
		ValidationResult valRes = (ValidationResult) adeComponent;
		assertEquals(ResultType.ERROR, valRes.getResult());

		ValidationError valError = valRes.getErrors().get(0);
		IntersectingRings error2 = (IntersectingRings) valError;
		assertEquals(error.getPolygonId(), error2.getPolygonId());
		assertEquals(error.getLinearRingId1(), error2.getLinearRingId1());
		assertEquals(error.getLinearRingId2(), error2.getLinearRingId2());
	}

	@Test
	void testWritingInteriorDisconnectedError()
			throws ADEException, CityGMLBuilderException, CityGMLWriteException, CityGMLReadException {
		CityModel model = new CityModel();
		Building b = new Building();
		model.getCityObjectMember().add(new CityObjectMember(b));

		ValidationResult result = new ValidationResult();
		result.setResult(ResultType.ERROR);
		b.addGenericApplicationPropertyOfCityObject(result);
		InteriorDisconnected error = new InteriorDisconnected();
		error.setPolygonId("TestPolygon");
		result.getErrors().add(error);

		byte[] gml = writeCityGml(model);
		model = readCityGml(gml);

		AbstractCityObject co = model.getCityObjectMember().get(0).getCityObject();
		List<ADEComponent> ades = co.getGenericApplicationPropertyOfCityObject();
		ADEComponent adeComponent = ades.get(0);
		ValidationResult valRes = (ValidationResult) adeComponent;
		assertEquals(ResultType.ERROR, valRes.getResult());

		ValidationError valError = valRes.getErrors().get(0);
		InteriorDisconnected error2 = (InteriorDisconnected) valError;
		assertEquals(error.getPolygonId(), error2.getPolygonId());
	}

	@Test
	void testWritingRingNotClosedError()
			throws ADEException, CityGMLBuilderException, CityGMLWriteException, CityGMLReadException {
		CityModel model = new CityModel();
		Building b = new Building();
		model.getCityObjectMember().add(new CityObjectMember(b));

		ValidationResult result = new ValidationResult();
		result.setResult(ResultType.ERROR);
		b.addGenericApplicationPropertyOfCityObject(result);
		RingNotClosed error = new RingNotClosed();
		error.setLinearRingId("TestRing");
		result.getErrors().add(error);

		byte[] gml = writeCityGml(model);
		model = readCityGml(gml);

		AbstractCityObject co = model.getCityObjectMember().get(0).getCityObject();
		List<ADEComponent> ades = co.getGenericApplicationPropertyOfCityObject();
		ADEComponent adeComponent = ades.get(0);
		ValidationResult valRes = (ValidationResult) adeComponent;
		assertEquals(ResultType.ERROR, valRes.getResult());

		ValidationError valError = valRes.getErrors().get(0);
		RingNotClosed error2 = (RingNotClosed) valError;
		assertEquals(error.getLinearRingId(), error2.getLinearRingId());
	}

	@Test
	void testWritingTooFewPointsError()
			throws ADEException, CityGMLBuilderException, CityGMLWriteException, CityGMLReadException {
		CityModel model = new CityModel();
		Building b = new Building();
		model.getCityObjectMember().add(new CityObjectMember(b));

		ValidationResult result = new ValidationResult();
		result.setResult(ResultType.ERROR);
		b.addGenericApplicationPropertyOfCityObject(result);
		TooFewPoints error = new TooFewPoints();
		error.setLinearRingId("TestRing");
		result.getErrors().add(error);

		byte[] gml = writeCityGml(model);
		model = readCityGml(gml);

		AbstractCityObject co = model.getCityObjectMember().get(0).getCityObject();
		List<ADEComponent> ades = co.getGenericApplicationPropertyOfCityObject();
		ADEComponent adeComponent = ades.get(0);
		ValidationResult valRes = (ValidationResult) adeComponent;
		assertEquals(ResultType.ERROR, valRes.getResult());

		ValidationError valError = valRes.getErrors().get(0);
		TooFewPoints error2 = (TooFewPoints) valError;
		assertEquals(error.getLinearRingId(), error2.getLinearRingId());
	}

	@Test
	void testWritingConsecutiveSameError()
			throws ADEException, CityGMLBuilderException, CityGMLWriteException, CityGMLReadException {
		CityModel model = new CityModel();
		Building b = new Building();
		model.getCityObjectMember().add(new CityObjectMember(b));

		ValidationResult result = new ValidationResult();
		result.setResult(ResultType.ERROR);
		b.addGenericApplicationPropertyOfCityObject(result);

		ConsecutivePointsSame error = new ConsecutivePointsSame();
		error.setLinearRingId("TestRing");
		DirectPosition vertex1 = new DirectPosition();
		vertex1.getValue().add(10.0);
		vertex1.getValue().add(5.0);
		vertex1.getValue().add(7.0);
		error.setVertex1(vertex1);
		DirectPosition vertex2 = new DirectPosition();
		vertex2.getValue().add(15.0);
		vertex2.getValue().add(55.0);
		vertex2.getValue().add(71.0);
		error.setVertex2(vertex2);
		result.getErrors().add(error);

		byte[] gml = writeCityGml(model);
		model = readCityGml(gml);

		AbstractCityObject co = model.getCityObjectMember().get(0).getCityObject();
		List<ADEComponent> ades = co.getGenericApplicationPropertyOfCityObject();
		ADEComponent adeComponent = ades.get(0);
		ValidationResult valRes = (ValidationResult) adeComponent;
		assertEquals(ResultType.ERROR, valRes.getResult());

		ValidationError valError = valRes.getErrors().get(0);
		ConsecutivePointsSame error2 = (ConsecutivePointsSame) valError;
		assertEquals(error.getLinearRingId(), error2.getLinearRingId());
		assertEquals(error.getVertex1().getValue().get(0), error2.getVertex1().getValue().get(0));
		assertEquals(error.getVertex2().getValue().get(0), error2.getVertex2().getValue().get(0));
	}

	@Test
	void testWritingRingSelfIntersection()
			throws ADEException, CityGMLBuilderException, CityGMLWriteException, CityGMLReadException {
		CityModel model = new CityModel();
		Building b = new Building();
		model.getCityObjectMember().add(new CityObjectMember(b));

		ValidationResult result = new ValidationResult();
		result.setResult(ResultType.ERROR);
		b.addGenericApplicationPropertyOfCityObject(result);

		RingSelfIntersection error = new RingSelfIntersection();
		error.setType(RingSelfIntType.DUPLICATE_POINT);
		error.setLinearRingId("TestRing");
		DirectPosition vertex1 = new DirectPosition();
		vertex1.getValue().add(10.0);
		vertex1.getValue().add(5.0);
		vertex1.getValue().add(7.0);
		error.setVertex1(vertex1);
		DirectPosition vertex2 = new DirectPosition();
		vertex2.getValue().add(15.0);
		vertex2.getValue().add(55.0);
		vertex2.getValue().add(71.0);
		error.setVertex2(vertex2);
		result.getErrors().add(error);

		byte[] gml = writeCityGml(model);
		model = readCityGml(gml);

		AbstractCityObject co = model.getCityObjectMember().get(0).getCityObject();
		List<ADEComponent> ades = co.getGenericApplicationPropertyOfCityObject();
		ADEComponent adeComponent = ades.get(0);
		ValidationResult valRes = (ValidationResult) adeComponent;
		assertEquals(ResultType.ERROR, valRes.getResult());

		ValidationError valError = valRes.getErrors().get(0);
		RingSelfIntersection error2 = (RingSelfIntersection) valError;
		assertEquals(error.getLinearRingId(), error2.getLinearRingId());
		assertEquals(error.getVertex1().getValue().get(0), error2.getVertex1().getValue().get(0));
		assertEquals(error.getVertex2().getValue().get(0), error2.getVertex2().getValue().get(0));
	}

	@Test
	void testWritingValidationPlan()
			throws ADEException, CityGMLBuilderException, CityGMLWriteException, CityGMLReadException {
		CityModel model = new CityModel();
		Validation val = new Validation();
		model.addGenericApplicationPropertyOfCityModel(val);

		ZonedDateTime expectedDate = LocalDateTime.of(2002, 5, 30, 9, 0).atZone(ZoneId.systemDefault());

		val.setValidationDate(expectedDate);
		val.setValidationSoftware("CityDoctor2");
		ValidationPlan plan = new ValidationPlan();
		val.setValidationPlan(plan);

		Parameter roundingParameter = new Parameter();
		roundingParameter.setName("numberOfRoundingPlaces");
		roundingParameter.setValue("8");
		plan.getGlobalParameters().add(roundingParameter);

		Parameter minVertexDistance = new Parameter();
		minVertexDistance.setName("minVertexDistance");
		minVertexDistance.setValue("0.0001");
		minVertexDistance.setUom("m");
		plan.getGlobalParameters().add(minVertexDistance);

		Requirement rTooFewPoints = new Requirement();
		rTooFewPoints.setEnabled(true);
		rTooFewPoints.setName(RequirementId.R_GE_R_TOO_FEW_POINTS);
		plan.getRequirements().add(rTooFewPoints);

		Checking buildingChecking = new Checking();
		buildingChecking.setValue(TopLevelFeatureType.BUILDING);
		plan.getFilter().add(buildingChecking);

		Statistics statistics = new Statistics();
		val.setStatistics(statistics);

		FeatureStatistics numErrorBridgeObjects = new FeatureStatistics();
		numErrorBridgeObjects.setNumChecked(0);
		numErrorBridgeObjects.setNumErrors(0);
		statistics.setNumErrorBridgeObjects(numErrorBridgeObjects);

		FeatureStatistics numErrorBuildings = new FeatureStatistics();
		numErrorBuildings.setNumChecked(1);
		numErrorBuildings.setNumErrors(1);
		statistics.setNumErrorBuildings(numErrorBuildings);

		FeatureStatistics numErrors = new FeatureStatistics();
		statistics.setNumErrorLandObjects(numErrors);
		statistics.setNumErrorTransportation(numErrors);
		statistics.setNumErrorVegetation(numErrors);
		statistics.setNumErrorWaterObjects(numErrors);

		ErrorStatistics stats1 = new ErrorStatistics();
		stats1.setAmount(1);
		stats1.setName(ErrorId.GE_R_CONSECUTIVE_POINTS_SAME);
		statistics.getErrorStatistics().add(stats1);

		byte[] gml = writeCityGml(model);
		model = readCityGml(gml);

		testValidationPlan(model);
	}

	private CityModel readCityGml(InputStream input)
			throws ADEException, CityGMLBuilderException, CityGMLReadException {
		CityGMLContext context = CityGMLContext.getInstance();
		List<ADEContext> adeContexts = context.getADEContexts();
		if (adeContexts == null || adeContexts.isEmpty()) {
			for (ADEContext adeContext : ServiceLoader.load(ADEContext.class)) {
				context.registerADEContext(adeContext);
			}
		}

		CityGMLBuilder builder = context.createCityGMLBuilder();
		CityGMLInputFactory in = builder.createCityGMLInputFactory();
		in.setProperty(CityGMLInputFactory.USE_VALIDATION, true);
		in.setProperty(CityGMLInputFactory.FEATURE_READ_MODE, FeatureReadMode.NO_SPLIT);
		in.setValidationEventHandler(event -> {
			System.out.print(
					"[" + event.getLocator().getLineNumber() + "," + event.getLocator().getColumnNumber() + "] ");
			System.out.println(event.getMessage());
			return true;
		});
		CityModel cityModel;
		try (CityGMLReader reader = in.createCityGMLReader("?", input)) {
			cityModel = (CityModel) reader.nextFeature();
		}
		return cityModel;
	}

	private CityModel readCityGml(File f)
			throws FileNotFoundException, ADEException, CityGMLBuilderException, CityGMLReadException {
		return readCityGml(new FileInputStream(f));
	}

	private CityModel readCityGml(String fileName)
			throws FileNotFoundException, ADEException, CityGMLBuilderException, CityGMLReadException {
		return readCityGml(new File(fileName));
	}

	private CityModel readCityGml(byte[] bytes) throws ADEException, CityGMLBuilderException, CityGMLReadException {
		return readCityGml(new ByteArrayInputStream(bytes));
	}

	private void testValidationPlan(CityModel cityModel) {
		assertFalse(cityModel.getGenericApplicationPropertyOfCityModel().isEmpty());
		List<ADEComponent> ades = cityModel.getGenericApplicationPropertyOfCityModel();
		ADEComponent adeComponent = ades.get(0);
		assertTrue(adeComponent instanceof Validation);
		Validation val = (Validation) adeComponent;
		ZonedDateTime expectedDate = LocalDateTime.of(2002, 5, 30, 9, 0).atZone(ZoneId.systemDefault());
		assertTrue(expectedDate.isEqual(val.getValidationDate()));
		assertEquals("CityDoctor2", val.getValidationSoftware());
		assertNotNull(val.getValidationPlan());
		ValidationPlan plan = val.getValidationPlan();
		assertNotNull(plan.getGlobalParameters());
		assertFalse(plan.getGlobalParameters().isEmpty());
		List<Parameter> globalParameters = plan.getGlobalParameters();
		Parameter roundingParameter = globalParameters.get(0);
		assertEquals("numberOfRoundingPlaces", roundingParameter.getName());
		assertEquals("8", roundingParameter.getValue());
		assertNull(roundingParameter.getUom());
		Parameter minVertexDistance = globalParameters.get(1);
		assertEquals("minVertexDistance", minVertexDistance.getName());
		assertEquals("0.0001", minVertexDistance.getValue());
		assertEquals("m", minVertexDistance.getUom());

		assertNotNull(plan.getRequirements());
		assertFalse(plan.getRequirements().isEmpty());

		List<Requirement> requirements = plan.getRequirements();
		Requirement rTooFewPoints = requirements.get(0);
		assertEquals(RequirementId.R_GE_R_TOO_FEW_POINTS, rTooFewPoints.getName());
		assertTrue(rTooFewPoints.isEnabled());

		assertNotNull(plan.getFilter());
		assertFalse(plan.getFilter().isEmpty());

		List<Checking> filter = plan.getFilter();
		Checking buildingChecking = filter.get(0);
		assertEquals(TopLevelFeatureType.BUILDING, buildingChecking.getValue());

		Statistics statistics = val.getStatistics();
		assertNotNull(statistics);
		FeatureStatistics numErrorBridgeObjects = statistics.getNumErrorBridgeObjects();
		assertNotNull(numErrorBridgeObjects);
		assertEquals(0, numErrorBridgeObjects.getNumChecked());
		assertEquals(0, numErrorBridgeObjects.getNumErrors());

		FeatureStatistics numErrorBuildings = statistics.getNumErrorBuildings();
		assertEquals(1, numErrorBuildings.getNumChecked());
		assertEquals(1, numErrorBuildings.getNumErrors());

		List<ErrorStatistics> errorStatistics = statistics.getErrorStatistics();
		assertNotNull(errorStatistics);
		assertFalse(errorStatistics.isEmpty());
		ErrorStatistics stats1 = errorStatistics.get(0);
		assertEquals(ErrorId.GE_R_CONSECUTIVE_POINTS_SAME, stats1.getName());
		assertEquals(1, stats1.getAmount());
	}

	private byte[] writeCityGml(CityModel model) throws ADEException, CityGMLBuilderException, CityGMLWriteException {
		CityGMLContext context = CityGMLContext.getInstance();
		List<ADEContext> adeContexts = context.getADEContexts();
		if (adeContexts == null || adeContexts.isEmpty()) {
			for (ADEContext adeContext : ServiceLoader.load(ADEContext.class)) {
				context.registerADEContext(adeContext);
			}
		}

		CityGMLBuilder builder = context.createCityGMLBuilder();

		CityGMLOutputFactory outputFactory = builder.createCityGMLOutputFactory(CityGMLVersion.DEFAULT);
		ByteArrayOutputStream out = new ByteArrayOutputStream();
		try (CityGMLWriter writer = outputFactory.createCityGMLWriter(out)) {
			writer.setIndentString("  ");
			writer.setPrefixes(CityGMLVersion.DEFAULT);
			writer.setPrefix("qual", QualityADEModule.NAMESPACE_URI);
			writer.setSchemaLocation(QualityADEModule.NAMESPACE_URI,
					"https://transfer.hft-stuttgart.de/pages/qualityade/0.1/qualityAde.xsd");
			writer.setSchemaLocations(CityGMLVersion.DEFAULT);
			writer.write(model);
			return out.toByteArray();
		}
	}

}
